@@ -8,18 +8,20 @@ package Petal;
 use Petal::Hash;
 use Petal::Cache::Disk;
 use Petal::Cache::Memory;
-use Petal::Parser::XMLWrapper;
-use Petal::Parser::HTMLWrapper;
+use Petal::Parser;
 use Petal::Canonicalizer::XML;
 use Petal::Canonicalizer::XHTML;
 use Petal::Functions;
 use Petal::Entities;
+use Petal::Encode;
 use File::Spec;
-use strict;
-use warnings;
 use Carp;
 use Safe;
 use Data::Dumper;
+use Scalar::Util;
+use strict;
+use warnings;
+use MKDoc::XML::Decode;
 
 
 # these are used as local variables when the XML::Parser
@@ -43,9 +45,9 @@ our $WARN_UNINIT = 0;
 # What do we use to parse input?
 our $INPUT  = 'XML';
 our $INPUTS = {
-    'XML'   => 'Petal::Parser::XMLWrapper',
-    'HTML'  => 'Petal::Parser::HTMLWrapper',
-    'XHTML' => 'Petal::Parser::HTMLWrapper',
+    'XML'     => 'Petal::Parser',
+    'HTML'    => 'Petal::Parser',
+    'XHTML'   => 'Petal::Parser',
 };
 
 
@@ -85,7 +87,7 @@ our $CURRENT_INCLUDES = 0;
 
 
 # this is for CPAN
-our $VERSION = '1.06';
+our $VERSION = '2.00';
 
 
 # The CodeGenerator class backend to use.
@@ -105,6 +107,10 @@ our $NS_URI = 'http://purl.org/petal/1.0/';
 our $XI_NS = 'xi';
 our $XI_NS_URI = 'http://www.w3.org/2001/XInclude';
 
+our $MT_NS       = 'metal';
+our $MT_NS_URI   = 'http://xml.zope.org/namespaces/metal';
+our $MT_NAME_CUR = 'main';
+
 
 # Displays the canonical template for template.xml.
 # You can set $INPUT using by setting the PETAL_INPUT environment variable.
@@ -145,14 +151,15 @@ sub main::lcode
 
 sub load_code_generator
 {
-	if (not $CodeGeneratorLoaded)
-	{
-	    eval "require $CodeGenerator";
-	    confess "Failed to load $CodeGenerator, $@" if $@;
-	    $CodeGeneratorLoaded = 1;
-	}
+    if (not $CodeGeneratorLoaded)
+    {
+	eval "require $CodeGenerator";
+	confess "Failed to load $CodeGenerator, $@" if $@;
+	$CodeGeneratorLoaded = 1;
+    }
 }
 
+
 # Instanciates a new Petal object.
 sub new
 {
@@ -170,15 +177,32 @@ sub new
 # if the language has been specified, let's try to
 # find which template we can use.
 sub _initialize
+{
+    my $self  = shift;
+    my $file  = $self->{file};
+    if ($file =~ /#/)
+    {
+	my ($file, $macro) = split /#/, $file, 2;
+	$self->{file}  = $file;
+	$self->_initialize_lang();
+	$self->{file} .= "#$macro";
+    }
+    else
+    {
+	$self->_initialize_lang();
+    }
+}
+
+
+sub _initialize_lang
 {
     my $self = shift;
     my $lang = $self->language() || return;
-    
     my @dirs = $self->base_dir();
-    @dirs = map { File::Spec->canonpath ("$_/$self->{file}") } @dirs;
+    @dirs    = map { File::Spec->canonpath ("$_/$self->{file}") } @dirs;
     
-    $self->{file} =~ s/\/$//;
-    my $filename = Petal::Functions::find_filename ($lang, @dirs);
+    $self->{file}  =~ s/\/$//;
+    my $filename   = Petal::Functions::find_filename ($lang, @dirs);
     $self->{file} .= "/$filename" if ($filename);
 }
 
@@ -238,6 +262,13 @@ sub _include_compute_path
 {
     my $self  = shift;
     my $file  = shift;
+    
+    # this is for metal self-includes
+    if ($file =~ /^#/)
+    {
+	$file = $self->{file} . $file;
+    }
+    
     return $file unless ($file =~ /^\./);
     
     my $path = $self->{file};
@@ -284,32 +315,10 @@ sub _include_compute_path
 sub process
 {
     my $self = shift;
+    $self->_process_absolutize_pathes();
     
-    # sets $BASE_DIR, @BASE_DIR and $self->{base_dir} to absolute paths once
-    # because File::Spec::rel2abs() can be quite expensive
-    $BASE_DIR = File::Spec->rel2abs ($BASE_DIR) if (defined $BASE_DIR and substr ($BASE_DIR, 0, 1) ne '/');
-    
-    @BASE_DIR = ( map { substr ($_, 0, 1) ne '/' ? File::Spec->rel2abs ($_) : $_ }
-		  map { defined $_ ? $_ : () } @BASE_DIR );
-
-    if ($self->{base_dir})
-    {
-	if (ref $self->{base_dir})
-	{
-	    $self->{base_dir} = [
-		map { substr ($_, 0, 1) ne '/' ? File::Spec->rel2abs ($_) : $_ }
-		map { defined $_ ? $_ : () } @{$self->{base_dir}}
-	       ] if (defined $self->{base_dir});
-	}
-	else
-	{
-	    $self->{base_dir} = File::Spec->rel2abs ($self->{base_dir}) unless (substr ($self->{base_dir}, 0, 1) ne '/');
-	}
-    }
-    
-    # ok, from there on we need to override any global
-    # variable with stuff that might have been specified
-    # when constructing the object
+    # ok, from there on we need to override any global variable with stuff
+    # that might have been specified when constructing the object
     local $TAINT              = defined $self->{taint}              ? $self->{taint}              : $TAINT;
     local $ERROR_ON_UNDEF_VAR = defined $self->{error_on_undef_var} ? $self->{error_on_undef_var} : $ERROR_ON_UNDEF_VAR;
     local $DISK_CACHE         = defined $self->{disk_cache}         ? $self->{disk_cache}         : $DISK_CACHE;
@@ -318,7 +327,7 @@ sub process
     local $INPUT              = defined $self->{input}              ? $self->{input}              : $INPUT;
     local $OUTPUT             = defined $self->{output}             ? $self->{output}             : $OUTPUT;
     local $BASE_DIR           = defined $self->{base_dir} ? do { ref $self->{base_dir} ? undef : $self->{base_dir} } : $BASE_DIR;
-    local @BASE_DIR           = defined $self->{base_dir} ? do { ref $self->{base_dir} ? @{$self->{base_dir}} : undef } : @BASE_DIR;
+    local @BASE_DIR           = defined $self->{base_dir} ? do { ref $self->{base_dir} ? @{$self->{base_dir}} : () } : @BASE_DIR;
     local $LANGUAGE           = defined $self->{default_language}   ? $self->{default_language}   : $LANGUAGE;
     local $DEBUG_DUMP         = defined $self->{debug_dump}         ? $self->{debug_dump}         : $DEBUG_DUMP;
     local $DECODE_CHARSET     = defined $self->{decode_charset}     ? $self->{decode_charset}     : $DECODE_CHARSET;
@@ -337,11 +346,12 @@ sub process
 	else                            { $hash = new Petal::Hash (@_)         }
 	
 	my $coderef = $self->_code_memory_cached;
+	die "\$coderef is undefined\n\n" unless $coderef;
+	die "\$hash is undefined\n\n" unless $hash;
 	$res = $coderef->($hash);
 	
 	$Petal::ENCODE_CHARSET and do {
-	    require "Encode.pm";
-	    $res = Encode::encode ($Petal::ENCODE_CHARSET, $res);
+	    $res = Petal::Encode::p_encode ($Petal::ENCODE_CHARSET, $res);
 	};
     };
     
@@ -350,6 +360,43 @@ sub process
 }
 
 
+# File::Spec->rel2abs() is pretty slow since it uses Cwd which does a
+# super-ugly backtick. Hence this method absolutizes base directories
+# only once. It is necessary to work with absolute base directories to
+# avoid cache conflicts.
+sub _process_absolutize_pathes
+{
+    my $self = shift;
+    
+    if (defined $BASE_DIR)
+    {
+	$BASE_DIR = File::Spec->rel2abs ($BASE_DIR) unless (
+	    File::Spec->file_name_is_absolute ($BASE_DIR)
+	     );
+    }
+    
+    @BASE_DIR = ( map { File::Spec->file_name_is_absolute ($_) ? $_ : File::Spec->rel2abs ($_) }
+		  map { defined $_ ? $_ : () } @BASE_DIR );
+    
+    if (defined $self->{base_dir})
+    {
+	if (ref $self->{base_dir})
+	{
+	    $self->{base_dir} = [
+		map { File::Spec->file_name_is_absolute ($_) ? $_ : File::Spec->rel2abs ($_) }
+		map { defined $_ ? $_ : () } @{$self->{base_dir}}
+	       ] if (defined $self->{base_dir});
+	}
+	else
+	{
+	    $self->{base_dir} = File::Spec->rel2abs ($self->{base_dir}) unless (
+		File::Spec->file_name_is_absolute ($self->{base_dir})
+		 );
+	}
+    }
+}
+
+
 sub _handle_error
 {
     my $self = shift;
@@ -408,16 +455,16 @@ sub _code_with_line_numbers
     # add line numbers
     my $count = 0;
     @lines = map {
-      my $cur_line = $_;
-      $count++;
-
-      # space padding so the line numbers nicely line up with each other
-      my $line_num = sprintf("%" . length(scalar(@lines)) . "d", $count);
-
-      # put line number and line back together
-      "${line_num}. ${cur_line}";
+	my $cur_line = $_;
+	$count++;
+	
+	# space padding so the line numbers nicely line up with each other
+	my $line_num = sprintf ("%" . length(scalar(@lines)) . "d", $count);
+	
+	# put line number and line back together
+	"${line_num}. ${cur_line}";
     } @lines;
-
+    
     return join("\n", @lines);
 }
 
@@ -434,6 +481,25 @@ sub _file
 }
 
 
+sub _macro
+{
+    my $self = shift;
+    my $file = $self->_file;
+    $file =~ s/^.*#// || return;
+    return $file;
+}
+
+
+sub _file_path_with_macro
+{
+    my $self  = shift;
+    my $file  = $self->_file_path;
+    my $macro = $self->_macro;
+    my $res   = $macro ? "$file#$macro" : $file;
+    return $res;
+}
+
+
 # $self->_file_path;
 # ------------------
 #   computes the file of the absolute path where the template
@@ -442,6 +508,7 @@ sub _file_path
 {
     my $self = shift;
     my $file = $self->_file;
+    $file =~ s/#.*$//;
     my @dirs = $self->base_dir;
     
     foreach my $dir (@dirs)
@@ -466,6 +533,7 @@ sub _file_data_ref
 {
     my $self      = shift;
     my $file_path = $self->_file_path;
+    $file_path =~ s/#.*$//;
     
     use bytes;
     open FP, "<$file_path" || die 'Cannot read-open $file_path';
@@ -473,83 +541,73 @@ sub _file_data_ref
     close FP;
     no bytes;
     
-    if ($] > 5.007 and $Petal::DECODE_CHARSET)
-    {
-	require "Encode.pm";
-	$res = Encode::decode ($Petal::DECODE_CHARSET, $res);
-    }
-    
-    if ($OUTPUT eq 'HTML' or $OUTPUT eq 'XHTML')
-    {
-	if ($] > 5.007)
-	{
-	    require "Encode.pm";
-	    Encode::_utf8_on ($res);
-	    Petal::Entities::decode_entities ($res);
-	    Encode::_utf8_off ($res);
-	}
-	else
-	{
-	    Petal::Entities::decode_entities ($res);
-	}	
-    }
+    $Petal::DECODE_CHARSET and do {
+	$res = Petal::Encode::p_decode ($Petal::DECODE_CHARSET, $res);
+    };
     
     # kill template comments
     $res =~ s/\<!--\?.*?\-->//gsm;
+    
+    my $decode = ($OUTPUT =~ /HTML$/i or $INPUT =~ /HTML$/i) ?
+        new MKDoc::XML::Decode ('numeric', 'xhtml') :
+	new MKDoc::XML::Decode ('numeric');
+    
+    Petal::Encode->p_utf8_on ($res);
+    $res = $decode->process ($res);
+    Petal::Encode->p_utf8_off ($res);
+    
     return \$res;
 }
 
 
 # $self->_code_disk_cached;
 # -------------------------
-#   Returns the Perl code data, using the disk cache if
-#   possible
+# Returns the Perl code data, using the disk cache if possible
 sub _code_disk_cached
 {
     my $self = shift;
-    my $file = $self->_file_path;
-    my $code = (defined $DISK_CACHE and $DISK_CACHE) ? Petal::Cache::Disk->get ($file) : undef;
+    my $code = (defined $DISK_CACHE and $DISK_CACHE) ? Petal::Cache::Disk->get ($self->_file_path_with_macro) : undef;
     unless (defined $code)
     {
-	my $data_ref = $self->_file_data_ref;
-	$data_ref    = $self->_canonicalize;
-
+	my $macro = $self->_macro() || $MT_NAME_CUR;
+	
+	local ($MT_NAME_CUR);
+	$MT_NAME_CUR = $macro;
+	
+	my $data_ref = $self->_canonicalize;
 	load_code_generator();
 	$code = $CodeGenerator->process ($data_ref, $self);
-	Petal::Cache::Disk->set ($file, $code) if (defined $DISK_CACHE and $DISK_CACHE);
+	Petal::Cache::Disk->set ($self->_file_path_with_macro, $code) if (defined $DISK_CACHE and $DISK_CACHE);
     }
+    
     return $code;
 }
 
 
 # $self->_code_memory_cached;
 # ---------------------------
-#   Returns the Perl code data, using the disk cache if
-#   possible
+# Returns the Perl code data, using the disk cache if possible
 sub _code_memory_cached
 {
     my $self = shift;
-    my $file = $self->_file_path;
-    my $code = (defined $MEMORY_CACHE and $MEMORY_CACHE) ? Petal::Cache::Memory->get ($file) : undef;
+    my $code = (defined $MEMORY_CACHE and $MEMORY_CACHE) ? Petal::Cache::Memory->get ($self->_file_path_with_macro) : undef;
     unless (defined $code)
     {
 	my $code_perl = $self->_code_disk_cached;
-        my $VAR1 = undef;
+	my $VAR1 = undef;
 	
 	if ($TAINT)
 	{
-	    # important line, don't remove
-	    ($code_perl) = $code_perl =~ m/^(.+)$/s;
-	    my $cpt = Safe->new ("Petal::CPT");
-	    $cpt->permit ('entereval');
-	    $cpt->permit ('leaveeval');
-	    $cpt->permit ('require');
-	    $cpt->reval($code_perl);
-	    confess ($@ . "\n" . $self->_code_with_line_numbers) if $@;
-	    
-	    # remove silly warning '"Petal::CPT::VAR1" used only once'
-	    $Petal::CPT::VAR1 if (0);
-	    $code = $Petal::CPT::VAR1;
+		# important line, don't remove
+		($code_perl) = $code_perl =~ m/^(.+)$/s;
+		die "\$code_perl is empty after untainting!" unless defined $code_perl && $code_perl;
+		my $cpt = Safe->new ("Petal::CPT");
+		$cpt->permit ('entereval');
+		$cpt->permit ('leaveeval');
+		$cpt->permit ('require');
+		$code = $cpt->reval($code_perl);
+		confess ("Error in reval:\n" . $@ . "\n" . $self->_code_with_line_numbers) if $@;
+		warn "\$code is empty after reval.\n" . Dumper($code, $Petal::CPT::VAR1, length($code_perl)) unless $code;
 	}
 	else
 	{
@@ -558,12 +616,16 @@ sub _code_memory_cached
 	    $code = $VAR1;
 	}
 	
-        Petal::Cache::Memory->set ($file, $code) if (defined $MEMORY_CACHE and $MEMORY_CACHE);
+	Petal::Cache::Memory->set ($self->_file_path_with_macro, $code) if (defined $MEMORY_CACHE and $MEMORY_CACHE);	
     }
+    
     return $code;
 }
 
 
+
+
+
 # $self->_code_cache;
 # -------------------
 #   Returns TRUE if this object uses the code cache, FALSE otherwise
@@ -593,11 +655,25 @@ sub _canonicalize
 
 1;
 
+
 =head1 NAME
 
 Petal - Perl Template Attribute Language - TAL for Perl!
 
 
+=head1 IMPORTANT NOTE 
+
+From version 2.00 onwards Petal *requires* that you use well-formed XML. This
+is because Petal now uses L<MKDoc::XML::TreeBuilder> rather than
+L<HTML::TreeBuilder> and L<XML::Parser>.
+
+In particular, this version of Petal *CAN* break backwards compatibility if you
+were using Petal's HTML mode will non well formed XHTML.
+
+If you still want to use broken XHTML, you can Petal 2.00 in conjunction with
+L<Petal::Parser::HTB> which has been created for this purpose.
+
+
 =head1 SYNOPSIS
 
 in your Perl code:
@@ -1231,6 +1307,81 @@ C<replace>, C<content>, C<omit-tag> and C<attributes> are NOT supported in
 conjunction with XIncludes.
 
 
+=head1 MORE INCLUDES: METAL
+
+Petal now supports a subset of the METAL specification, which is a very WYSIWYG
+compatible way of doing includes. The current implementation supports only two
+metal statements: define-macro and use-macro.
+
+
+=head2 define-macro
+
+In order to define a macro inside a file (i.e. a fragment to be included), you
+use the metal:use-macro directive. For example:
+
+  File foo.xml
+  ============
+
+  <html>
+    <body>
+      <p metal:define-macro="footer">
+        (c) Me (r)(tm) (pouet pouet)
+      </p>
+    </body>
+  </html>
+
+
+=head2 use-macro
+
+In order to use a previously defined macro, you use the metal:define-macro directive.
+For example:
+
+  File bar.xml
+  ============
+
+  <html>
+    <body>
+      ... plenty of content ...
+
+      <p metal:use-macro="foo.xml#footer">
+        Page Footer.
+      </p>
+    </body>
+  </html>
+
+
+=head2 self includes
+
+In Zope, METAL macros are expanded first, and then the TAL instructions are processed.
+However with Petal, METAL macros are expanded at run-time just like regular includes,
+which allows for recursive macros.
+
+This example templates a sitemap, which on a hierarchically organized site would
+be recursive by nature:
+
+  <html>
+    <body>
+      <p>Sitemap:</p>
+
+      <li metal:define-macro="recurse">
+        <a href="#"
+           petal:attributes="href child/Full_Path" 
+           petal:content="child/Title"
+        >Child Document Title</a>
+        <ul 
+          petal:define="children child/Children"
+          petal:condition="children"
+          petal:repeat="child children"
+        >
+          <li metal:use-macro="#recurse">Dummy Child 1</li>
+          <li petal:replace="nothing">Dummy Child 2</li>
+          <li petal:replace="nothing">Dummy Child 3</li>
+        </ul>
+      </li>
+    </body>
+  </html>
+
+
 =head1 EXPRESSIONS AND MODIFIERS
 
 Petal has the ability to bind template variables to the following Perl
